#include <dvdread/ifo_read.h>
#include <stdio.h>
#include <unistd.h>

static struct { char code[3];	char name[20];}
language[] = {
	{ "  ", "Not Specified" }, { "aa", "Afar" },	{ "ab", "Abkhazian" }, { "af", "Afrikaans" },	{ "am", "Amharic" },
	{ "ar", "Arabic" }, { "as", "Assamese" },	{ "ay", "Aymara" }, { "az", "Azerbaijani" }, { "ba", "Bashkir" },
	{ "be", "Byelorussian" }, { "bg", "Bulgarian" }, { "bh", "Bihari" }, { "bi", "Bislama" }, { "bn", "Bengali; Bangla" },
	{ "bo", "Tibetan" }, { "br", "Breton" }, { "ca", "Catalan" }, { "co", "Corsican" }, { "cs", "Czech" },
	{ "cy", "Welsh" }, { "da", "Dansk" }, { "de", "Deutsch" }, { "dz", "Bhutani" }, { "el", "Greek" }, { "en", "English" },
	{ "eo", "Esperanto" }, { "es", "Espanol" }, { "et", "Estonian" }, { "eu", "Basque" }, { "fa", "Persian" },
	{ "fi", "Suomi" }, { "fj", "Fiji" }, { "fo", "Faroese" }, { "fr", "Francais" }, { "fy", "Frisian" }, { "ga", "Gaelic" },
	{ "gd", "Scots Gaelic" }, { "gl", "Galician" }, { "gn", "Guarani" }, { "gu", "Gujarati" }, { "ha", "Hausa" },
	{ "he", "Hebrew" }, { "hi", "Hindi" }, { "hr", "Hrvatski" }, { "hu", "Magyar" }, { "hy", "Armenian" },
	{ "ia", "Interlingua" }, { "id", "Indonesian" }, { "ie", "Interlingue" }, { "ik", "Inupiak" }, { "in", "Indonesian" },     
	{ "is", "Islenska" }, { "it", "Italiano" }, { "iu", "Inuktitut" }, { "iw", "Hebrew" }, { "ja", "Japanese" },
	{ "ji", "Yiddish" }, { "jw", "Javanese" }, { "ka", "Georgian" }, { "kk", "Kazakh" }, { "kl", "Greenlandic" },
	{ "km", "Cambodian" }, { "kn", "Kannada" }, { "ko", "Korean" }, { "ks", "Kashmiri" }, { "ku", "Kurdish" },
	{ "ky", "Kirghiz" }, { "la", "Latin" }, { "ln", "Lingala" }, { "lo", "Laothian" }, { "lt", "Lithuanian" },
	{ "lv", "Latvian, Lettish" }, { "mg", "Malagasy" }, { "mi", "Maori" }, { "mk", "Macedonian" }, { "ml", "Malayalam" },
	{ "mn", "Mongolian" }, { "mo", "Moldavian" }, { "mr", "Marathi" }, { "ms", "Malay" }, { "mt", "Maltese" },
	{ "my", "Burmese" }, { "na", "Nauru" }, { "ne", "Nepali" }, { "nl", "Nederlands" }, { "no", "Norsk" }, { "oc", "Occitan" },
	{ "om", "Oromo" }, { "or", "Oriya" }, { "pa", "Punjabi" }, { "pl", "Polish" }, { "ps", "Pashto, Pushto" },
	{ "pt", "Portugues" }, { "qu", "Quechua" }, { "rm", "Rhaeto-Romance" }, { "rn", "Kirundi" }, { "ro", "Romanian"  },
	{ "ru", "Russian" }, { "rw", "Kinyarwanda" }, { "sa", "Sanskrit" }, { "sd", "Sindhi" }, { "sg", "Sangho" },
	{ "sh", "Serbo-Croatian" }, { "si", "Sinhalese" }, { "sk", "Slovak" }, { "sl", "Slovenian" }, { "sm", "Samoan" },
 	{ "sn", "Shona"  }, { "so", "Somali" }, { "sq", "Albanian" }, { "sr", "Serbian" }, { "ss", "Siswati" },
	{ "st", "Sesotho" }, { "su", "Sundanese" }, { "sv", "Svenska" }, { "sw", "Swahili" }, { "ta", "Tamil" },
	{ "te", "Telugu" }, { "tg", "Tajik" }, { "th", "Thai" }, { "ti", "Tigrinya" }, { "tk", "Turkmen" }, { "tl", "Tagalog" },
	{ "tn", "Setswana" }, { "to", "Tonga" }, { "tr", "Turkish" }, { "ts", "Tsonga" }, { "tt", "Tatar" }, { "tw", "Twi" },
	{ "ug", "Uighur" }, { "uk", "Ukrainian" }, { "ur", "Urdu" }, { "uz", "Uzbek" }, { "vi", "Vietnamese" },
	{ "vo", "Volapuk" }, { "wo", "Wolof" }, { "xh", "Xhosa" }, { "yi", "Yiddish" }, { "yo", "Yoruba" }, { "za", "Zhuang" },
	{ "zh", "Chinese" }, { "zu", "Zulu" }, { "xx", "Unknown" }, { "\0", "Unknown" } };
char *video_format[2] = {"NTSC", "PAL"};
char *aspect_ratio[4] = {"4/3", " ?:?", " ?:?", "16/9"};
char *ar2[4]          = {"1.33", " ???", " ???", "1.78"};
char *quantization[4] = {"16bit", "20bit", "24bit", "drc"};
char *mpeg_version[2] = {"mpeg1", "mpeg2"};
char *video_height[2] = {"480", "576"};
char *video_width[4] = {"720", "704", "352", "352"};
char *permitted_df[4] = {"P&S + Letter", "Pan&Scan", "Letterbox", "?"};
char *audio_format[7] = {"ac3", "?", "mpeg1", "mpeg2", "lpcm ", "sdds ", "dts"};
char *sample_freq[2]  = {"48kHz", "??kHz"};
char *audio_type[5]   = {"Undefined", "Normal", "Impaired", "Comments1", "Comments2"};
char *subp_type[16]   = {"Undefined", "Normal", "Large", "Children", "reserved", "Normal_CC", "Large_CC", "Children_CC", 
	"reserved", "Forced", "reserved", "reserved", "reserved", "Director", "Large_Director", "Children_Director"};
char *frames_per_s[4] = {"", "25.00", "", "29.97"};

char* program_name;

/*
 *  The following method is based on code from vobcopy, by Robos, with thanks.
 */

int get_title_name(const char* dvd_device, char* title)
{
  int filehandle = 0;
  int last_char = 0, blah = 0;
  int bytes_read;
  
  if (! (filehandle = open(dvd_device, 00))) {
    fprintf(stderr, "Couldn't open %s for title\n", dvd_device);
    title = "unknown";
    return -1;
  }
 
  if ( 32808 != lseek(filehandle, 32808, SEEK_SET )) {
    close(filehandle);
    fprintf(stderr, "Couldn't seek in %s for title\n", dvd_device);
    title = "unknown";
    return -1;   
  } 

  if ( 32 != ( read(filehandle, title, 32))) {
    close(filehandle);
    fprintf(stderr, "Couldn't read enough bytes for title. only read $d\n", bytes_read);
    title = "unknown";
    return -1;
  }

  close (filehandle);

  title[32] = '\0';
  for ( blah = 0; blah < 32; blah++ )
  { if ( title[blah] != ' ' ) { last_char = blah; } } 
  title[last_char+1] = '\0';

  return 0;
}

char* lang_name(char* code)
{
	int k=0;
	while (memcmp(language[k].code, code, 2) && language[k].name[0] ) { k++; }
	return language[k].name;
}

usage()
{
  fprintf(stderr, "Usage: %s [options] [-t track_number] [dvd path] \n\t Options: extra information on:\n", program_name);
  fprintf(stderr, "\t-a audio streams\n\t-c cells\n\t-n angles\n\t-p programs (chapters)\n\t-s subpictures\n");
  fprintf(stderr, "\t-v video\n\t-x all information\n\t-q quiet - no summary totals\n");
  fprintf(stderr, "\"Written\" by Chris Phillips <acid_kewpie@users.sourceforge.net>\n");
}

main(int argc, char *argv[])
{
        char title[35];
	dvd_reader_t *dvd;
	ifo_handle_t *ifo_zero, **ifo;
	dvd_time_t *dvd_time;
	pgcit_t *vts_pgcit;
	vtsi_mat_t *vtsi_mat;
	audio_attr_t *audio_attr;
	video_attr_t *video_attr;
	subp_attr_t *subp_attr;
	pgc_t *pgc;
	int i, j, c, titles, cell, vts_ttn, title_set_nr;
 	char lang_code[2];
	char *dvd_device = "/dev/dvd";

	int opt_a=0, opt_c=0, opt_n=0, opt_p=0, opt_q=0, opt_s=0, opt_t=0, opt_v=0, opt_x=0;

	program_name = argv[0];

	while ((c = getopt(argc, argv, "acnpqsvt:xh?")) != EOF) {
		switch (c) {
    		case 'h':
		case '?':	usage();	return 0;
		case 'a':	opt_a = 1;	break;
		case 'c':	opt_c = 1;	break;
		case 's':	opt_s = 1;	break;
		case 'q':	opt_q = 1;	break;
		case 'p':	opt_p = 1;	break;
		case 'n':	opt_n = 1;	break;
		case 't':	opt_t = atoi(optarg);	break;
		case 'v':	opt_v = 1;	break;
		case 'x':	opt_a = 1;
				opt_c = 1;      
                                opt_s = 1;      
                                opt_p = 1;      
                                opt_n = 1;      
                                opt_v = 1;	break;
		}
	}

	if (argv[optind]) { dvd_device = argv[optind];	}
	dvd = DVDOpen(dvd_device);
	if( !dvd ) {
		fprintf( stderr, "Can't open disc %s!\n", dvd_device );
		return 1;
	}
	ifo_zero = ifoOpen(dvd, 0);
	
	ifo = (ifo_handle_t **)malloc((ifo_zero->vts_atrt->nr_of_vtss + 1) * sizeof(ifo_handle_t *));	
	
	for (i=1; i <= ifo_zero->vts_atrt->nr_of_vtss; i++)	ifo[i] = ifoOpen(dvd, i);
	
	titles = ifo_zero->tt_srpt->nr_of_srpts;

	if ( opt_t > titles || opt_t < 0) { printf ("Only %d titles on this disc!", titles); return 1;}

        get_title_name(dvd_device, title);
	printf("Disc Title: %s\n", title);

	for (j=0; j < titles; j++)
	{
	if ( opt_t == j+1 || opt_t == 0 ) {
		// GENERAL 
		if (ifo[ifo_zero->tt_srpt->title[j].title_set_nr]->vtsi_mat) {
	
		vtsi_mat   = ifo[ifo_zero->tt_srpt->title[j].title_set_nr]->vtsi_mat;
		vts_pgcit  = ifo[ifo_zero->tt_srpt->title[j].title_set_nr]->vts_pgcit;
		video_attr = &vtsi_mat->vts_video_attr;
		vts_ttn = ifo_zero->tt_srpt->title[j].vts_ttn;
		title_set_nr = ifo_zero->tt_srpt->title[j].title_set_nr;
		//pgc = vts_pgcit->pgci_srp[ifo_zero->tt_srpt->title[j].vts_ttn - 1].pgc;
		//printf ("this pgcn = %i", ifo[title_set_nr]->vts_ptt_srpt->title[vts_ttn - 1].ptt[0].pgcn - 1);
		pgc = vts_pgcit->pgci_srp[ifo[title_set_nr]->vts_ptt_srpt->title[vts_ttn - 1].ptt[0].pgcn - 1].pgc;
		
		dvd_time = &pgc->playback_time;

		printf("Title: %02d, Length: %02x:%02x:%02x ", j+1, dvd_time->hour, dvd_time->minute, dvd_time->second);

		if (! opt_q) {
		printf("Chapters: %02d, Cells: %02d, ", ifo_zero->tt_srpt->title[j].nr_of_ptts, pgc->nr_of_cells);
		printf("Audio streams: %02d, Subpictures: %02d", vtsi_mat->nr_of_vts_audio_streams, vtsi_mat->nr_of_vts_subp_streams);
		}
		printf("\n");


		if (opt_v) {
		printf("\tVTS: %02d, TTN: %02d, ", ifo_zero->tt_srpt->title[j].title_set_nr, ifo_zero->tt_srpt->title[j].vts_ttn);
		printf("FPS: %s, ", frames_per_s[(dvd_time->frame_u & 0xc0) >> 6]);
		printf("Format: %s, Aspect ratio: %s, ", video_format[video_attr->video_format], aspect_ratio[video_attr->display_aspect_ratio]);
		printf("Width: %s, Height: %s, ", video_width[video_attr->picture_size], video_height[video_attr->video_format]);
		printf("DF: %s\n", permitted_df[video_attr->permitted_df]);
		}

		
		// ANGLES
		if (opt_n) {
		printf("\tNumber of Angles: %d\n", ifo_zero->tt_srpt->title[j].nr_of_angles);
		}

		// AUDIO	
		if (opt_a) {
		for (i=0; i<vtsi_mat->nr_of_vts_audio_streams; i++)
		{
			audio_attr = &vtsi_mat->vts_audio_attr[i];
			sprintf(lang_code, "%c%c", audio_attr->lang_code>>8, audio_attr->lang_code & 0xff);
			if (!lang_code[0]) { lang_code[0] = 'x'; lang_code[1] = 'x'; }
			printf("\tAudio: %d, Language: %s - %s, ", i +1 , lang_code, lang_name(lang_code));
			printf("Format: %s, ", audio_format[audio_attr->audio_format]);
			printf("Frequency: %s, ", sample_freq[audio_attr->sample_frequency]);
			printf("Quantization: %s, ", quantization[audio_attr->quantization]);
			printf("Channels: %d, AP: %d, ", audio_attr->channels+1, audio_attr->application_mode);
			printf("Content: %s", audio_type[audio_attr->lang_extension]);
			printf("\n");
		}
		}

		// CHAPTERS
		cell = 0;
		if (opt_p) {
		for (i=0; i<pgc->nr_of_programs; i++)
		{
			int second=0, minute=0, hour=0, tmp;
			char hex[2];
			int next = pgc->program_map[i+1];
			if (i == pgc->nr_of_programs - 1) next = pgc->nr_of_cells + 1;
			while (cell < next - 1)
			{
				dvd_time = &pgc->cell_playback[cell].playback_time;
				sprintf(hex, "%02x", dvd_time->second);
				tmp = second + atoi(hex);
				//tmp = 10 * (dvd_time->second >> 4) + (dvd_time->second & 0x0F); nah.. didn't work
				minute = minute + (tmp / 60);
				second = tmp % 60;
				sprintf(hex, "%02x", dvd_time->minute);
				tmp = minute + atoi(hex);
				hour = hour + (tmp / 60);
				minute = tmp % 60;
				cell++;
			}
			printf("\tChapter: %02d, Length: %02d:%02d:%02d\n", i+1, hour, minute, second);
		}
		}
		
		// CELLS
		//printf("\tNumber of Cells: %d\n", pgc->nr_of_cells);
		if (opt_c) {
		for (i=0; i<pgc->nr_of_cells; i++)
		{
			dvd_time = &pgc->cell_playback[i].playback_time;
			printf("\tCell: %02d, Length: %02x:%02x:%02x", i+1, dvd_time->hour, dvd_time->minute, dvd_time->second);
			printf("\n");
		}
		}
		
		// SUBTITLES
		if (opt_s) {
		for (i=0; i<vtsi_mat->nr_of_vts_subp_streams; i++)
		{
			//nullstring = "";
			subp_attr = &vtsi_mat->vts_subp_attr[i];
			sprintf(lang_code, "%c%c", subp_attr->lang_code>>8, subp_attr->lang_code & 0xff);
			if (!lang_code[0]) { lang_code[0] = 'x'; lang_code[1] = 'x'; }
			printf("\tSubtitle: %02d, Language: %s - %s, ", i+1, lang_code, lang_name(lang_code));
			printf("Content: %s", subp_type[subp_attr->lang_extension]);
			printf("\n");
		}
		}
			
		if ( opt_a || opt_c || opt_p || opt_s || opt_n) printf("\n");
	}
	}
}
        for (i=1; i <= ifo_zero->vts_atrt->nr_of_vtss; i++) {
		ifoClose(ifo[i]);
	}
	ifoClose(ifo_zero);
	DVDClose(dvd);

	return 0;
}
